/*
 * Sample code for SAS Global Forum Paper SAS2125-2018
 * The paper is available at:
 *   http://support.sas.com/resources/papers/proceedings18
 *
 * The module defines all the custom FCMP functions and 
 * subroutines used in the SAS code. As such, this file needs
 * to be run only once prior to the SAS code in the 
 * fcmpSasCode.sas file.
 * 
 * Author:      Bill McNeill
 * Company:     SAS Institute Inc.
 * E-mail:      Bill.McNeill@sas.com
 * Date:        22 February 2018
 *
 * Depending on which examples you want to run, here are the 
 * minimum version of SAS:
 *   Minimum SAS Version to use FCMP procedure:   9.2
 *   Minimum SAS Version to use STATIC statement: 9.4 maintenance 3
 *   Minimum SAS Version to use LEAVE statement:  9.4 maintenance 4
 */

   /* The directory specified in the LIBNAME statement must exist.  
      The same directory must be specified in the LIBNAME statement 
      used in the fcmpSasCode.sas file.                             */ 
   LIBNAME billsLib 'c:/SAS_Global_Forum/2018/sasuser/';

   /* Taxes */
   PROC FCMP OUTLIB=billsLib.paper.tax;
      /* Return the price with tax of the passed in value. 
         The tax rate is 7.25%.                            */
      FUNCTION priceWithTax_func(price);
         STATIC taxRate;
         IF (MISSING(taxRate))
           THEN taxRate = (7.25 / 100);
         taxedPrice = (price * (1 + taxRate));
         RETURN (taxedPrice);
         ENDSUB;

      /* Pass back in the taxedPrice variable the value 
         of the price variable with tax applied.        
         The tax rate is 7.25%.                         */
      SUBROUTINE priceWithTax_sub(price, taxedPrice);
         OUTARGS taxedPrice;
         STATIC taxRate;
         IF (MISSING(taxRate))
           THEN taxRate = (7.25 / 100);
         taxedPrice = (price * (1 + taxRate));
         ENDSUB;
      RUN;


   /* Vowels */
   PROC FCMP OUTLIB=billsLib.paper.vowels;
     /* Return the number of specified vowels in the 
        passed in string value.                      */
     FUNCTION getVowelCount (string $);
       ARRAY vowel[5] $ ('a', 'e', 'i', 'o', 'u');
 
       vowelCount = 0;
       DO stringIndex=1 to LENGTH(string);
          stringChar = SUBSTR(string, stringIndex, 1);
          DO vowelIndex=1 TO dim(vowel);
             IF ( (stringChar EQ        vowel[vowelIndex])  OR
                  (stringChar EQ UPCASE(vowel[vowelIndex]))    )
               THEN DO
                 vowelCount = vowelCount + 1;
                 /* Found a vowel, no need to continue  */
                 /* checking this character.            */
                 LEAVE;
                 END;
             END;
          END;
       RETURN(vowelCount);
       ENDSUB;

     FUNCTION getVowelCountStmt (product $) $;
       /* Return a string comprising a sentence noting the number of 
          vowels in the string passed into the function.             
	      Note that this function calls the previously defined       
	      getVowelCount function.                                    */
       LENGTH vowelStmt $60;
       vowelClause1 = "There are ";
       vowelCount   = getVowelCount(product);
       vowelClause2 = " vowels in the product. ";
       vowelStmt    = CAT(vowelClause1, vowelCount, vowelClause2, product);
       RETURN(vowelStmt);
       ENDSUB;
   RUN;


   /* Create a temporary function definition for demonstrating a 
      subroutine deletion from a package.                        */
   PROC FCMP OUTLIB=billsLib.paper.delete;
      FUNCTION priceWithTax_test1(price);
         STATIC taxRate;
         IF (MISSING(taxRate))
           THEN taxRate = (7.25 / 100);
         taxedPrice = (price * (1 + taxRate));
         RETURN (taxedPrice);
         ENDSUB;

      SUBROUTINE priceWithTax_test2(price, taxedPrice);
         OUTARGS taxedPrice;
         taxRate = (7.25 / 100);
         taxedPrice = (price * (1 + taxRate));
         ENDSUB;
      RUN;


   /* Tracing */
   PROC FCMP INLIB=billsLib.paper TRACE;
     ATTRIB itemCost FORMAT=DOLLAR8.2
            taxPrice FORMAT=DOLLAR8.2; 
     itemCost=6.99;
     taxPrice=priceWithTax_func(itemCost);
     PUT itemCost "will be " taxPrice " with tax";
     RUN;

   /* Temperature averages */
   PROC FCMP OUTLIB=billsLib.paper.averages;
     /* Determine the average temperature for the same time index
        over a month of specified daily temperature data.         */
     FUNCTION averageTempAtTime(dsName $, timeIndex);
        ARRAY monthlyTimeTemp[1,1] / NOSYMBOLS;
        /* Read the data from the named SAS data set. I found out 
		   after the paper was finalized that READ_ARRAY does 
		   dynamic sizing. Therefore, the array passed in can be 
		   declared as [1,1]                                      */
        rc = READ_ARRAY(dsName, monthlyTimeTemp);
        IF (rc EQ 0)
          THEN DO;
            averageTemp = getAverageTemp(monthlyTimeTemp, timeIndex);
            END;
          ELSE DO;
            /* Problem reading the data set. */
            ATTRIB missingValue FORMAT=8.;
            missingValue = .;
            averageTemp  = missingValue;
            END;
        RETURN(averageTemp);
        ENDSUB;

     /* Compute the average temperature for the specified time index. */
     FUNCTION getAverageTemp(monthlyTimeTemp[*,*], timeIndex);
	    /* Determine the size of the array being passed into this 
	       function. Note that the rows in the temperature data can
	       vary from 28 to 31 (maximum number of days in the months). */
        ARRAY tempByTime[1,1]  / NOSYMBOLS;
        rowDimension    = 1;
        columnDimension = 2;
        daysInMonth     = DIM(monthlyTimeTemp,rowDimension);
        readingsPerDay  = DIM(monthlyTimeTemp,columnDimension);

		/* Create a local array sized to match the size of the array 
		   passed it. This local array will hold the transposed data. */
        CALL DYNAMIC_ARRAY(tempByTime, readingsPerDay, daysInMonth);
		/* Transpose the data so that each row in the local array is 
		   the time when the temperature was taken and the columns 
		   are the days in the month.                                 */
        CALL TRANSPOSE(monthlyTimeTemp, tempByTime);
        timeSum = getArrayRowSum(tempByTime, timeIndex);
        averageTimeTemp = (timeSum / daysInMonth);
        RETURN(averageTimeTemp);
        ENDSUB;

     /* Get the sum of the numbers in a specified row of an array.   */
     FUNCTION getArrayRowSum(arrayToSum[*,*], rowIndex);
        columnDimension = 2;
        columns = DIM(arrayToSum, columnDimension);
        sum = 0;
        DO columnIndex=1 to columns;
           sum = (sum + arrayToSum[rowIndex, columnIndex]);
           END;
        RETURN(sum);
        ENDSUB;
      RUN;

   /* Run Macro */
   PROC FCMP OUTLIB=billsLib.paper.runMacro;
      /* Wrapper function to call the SAS macro code with the FCMP 
         special function RUN_MACRO to create the JSON file with 
         the data for the specified age.                             */
      FUNCTION writeJsonFile(dsName $, age);
         procRc = 0;
		 /* Call the SAS macro named "jsonFileCreate". The rest of
		    the arguments in the run_macro function are converted 
		    to macro variables.                                      */
         rc = run_macro("jsonFileCreate", dsName, age, procRc); 
		 /* Error checking                                           */
         IF (rc NE 0)
           THEN DO;
             PUT "Problems with submitting JSON file macro (" rc ")";
             END;
           ELSE DO;
             IF (procRc NE 0)
               THEN DO;
                 PUT "Problems with writing JSON file (" procRc ")";
                 rc = procRc;
                 END;
             END;
         RETURN(rc);
         ENDSUB;
      RUN;

   /* Recursion (FCMP function */
   PROC FCMP OUTLIB=billsLib.paper.factorial;
      /* Classic recursive example of computing a factorial of a 
         non-negative number. 
         Note that this example is for demonstration purposes. 
         SAS has a FACT function to do this calculation.             */
      FUNCTION factorial(currentValue);
        IF (currentValue EQ 0)
        THEN RETURN(1);
        ELSE RETURN(currentValue * factorial(currentValue - 1));
        ENDSUB;
      RUN;

   /* Recursion (external C function) 
      NOTE: The PROTO and FCMP procedures need to be able to 
      locate the image containing the C defined factorial 
      function. If not, error messages will appear in the
      SAS log.                                                       */ 
   PROC PROTO PACKAGE=billsLib.proto.factFunc;
      /* The name of the shared library containing the "factorial" 
         function written in C.                                      */
      LINK "factorial";
	    /* The C function prototype                                  */
        int factorial(int currentValue);
      RUN;

   PROC FCMP INLIB=billsLib.proto             
             OUTLIB=billsLib.paper.factorial;
	  /* FCMP wrapper function to call the C function to compute 
	     the factorial value of a number.                            */
      FUNCTION factorialProto(currentValue);
         RETURN(factorial(currentValue));
         ENDSUB;
      RUN;

   /* Hash Object */
   PROC FCMP OUTLIB=billsLib.paper.cars;
     /* Use the Hash object to replace the IN operator of a  
        WHERE statement.                                             */
     FUNCTION isPossibleCar(model $);
	    true  = 1;
		false = 0;
	    /* Declare and load the hash object with the data set 
	       "work.possCars". Note that the declare of the hash 
	       object and the calls to the "definekey" and "definedone" 
	       methods are only done the first time the function is 
	       called.                                                   */
        DECLARE HASH hashObj(DATASET:"work.possCars"); 
        /* Identify the "model" variable in the data set as the key 
		   value for the hash object.                                */
        rc = hashObj.definekey("model");
		/* The hash object definition is complete.                   */
        rc = hashObj.definedone();
		/* Hash object method to determine if the "model" value 
		   passed into this function is present in the hash object 
		   loaded with the "work.possCars" data set.                 */
        rc = hashObj.check();
		/* Convert the result of calling the "check" method to a 
		   Boolean value that can be used by the WHERE statement.    */
        IF (rc EQ 0)
          THEN isMatch = true;
          ELSE isMatch = false;
        RETURN(isMatch);  
        ENDSUB;
      RUN;
