/*
 * Sample code for SAS Global Forum Paper SAS2125-2018
 * The paper is available at:
 *   http://support.sas.com/resources/papers/proceedings18
 *
 * The module contains SAS code that demonstrates using the custom 
 * FCMP functions and subroutines defined by running the 
 * fcmpDefinition.sas file. As such, this file can only be run
 * after running the SAS code in the fcmpDefinition.sas file.
 * 
 * Author:      Bill McNeill
 * Company:     SAS Institute Inc.
 * E-mail:      Bill.McNeill@sas.com
 * Date:        22 February 2018
 *
 * Depending on which examples you want to run, here are the 
 * minimum version of SAS:
 *   Minimum SAS Version to use FCMP procedure:   9.2
 *   Minimum SAS Version to use STATIC statement: 9.4 maintenance 3
 *   Minimum SAS Version to use LEAVE statement:  9.4 maintenance 4
 */

   /* The directory specified in the LIBNAME statement must exist.  
      The same directory must be specified in the LIBNAME statement 
      used in the fcmpSasCode.sas file.                             */
   LIBNAME billsLib 'c:/SAS_Global_Forum/2018/sasuser/';

   /* The CMPLIB option identifies the data set containing the FCMP 
      definitions. This must match the value specified in the 
      OUTLIB= option used in the FCMP procedure defining the needed 
      FCMP function or subroutine.                                   */
   OPTION CMPLIB=billsLib.paper;

   /* TAXES (DATA step) */
   DATA work.carPriceWithTax(DROP=taxRate);
      /* Compute the price of the cars with the specified tax rate.  */
      SET sashelp.cars;
      FORMAT taxedPrice DOLLAR8.;
      RETAIN taxRate;
      IF (MISSING(taxRate))
        THEN taxRate = (7.25 / 100);
      taxedPrice = (msrp * (1 + taxRate));
      OUTPUT;
      RUN;

   PROC PRINT DATA=work.carPriceWithTax(
              KEEP=make model invoice msrp taxedPrice
              OBS=5); 
      RUN;

   /* TAXES (function) */
   DATA work.carPriceWithTaxFun;
      /* Compute the price of the cars by calling the 
         priceWithTax_func FCMP function.                            */
      SET sashelp.cars;
      FORMAT msrpWithTax DOLLAR8.;
      msrpWithTax = priceWithTax_func(msrp);
      OUTPUT;
      RUN;

   PROC PRINT DATA=work.carPriceWithTaxFun(
              KEEP=make model invoice msrp msrpWithTax
              OBS=5); 
      RUN;

   /* TAXES (subroutine) */
   DATA work.carPriceWithTaxSub;
      /* Compute the price of the cars by calling the 
         priceWithTax_sub FCMP subroutine.                           */
      SET sashelp.cars;
      FORMAT msrpWithTax DOLLAR8.;
      msrpWithTax = 0;
      CALL priceWithTax_sub(msrp, msrpWithTax);
      OUTPUT;
      RUN;

   PROC PRINT DATA=work.carPriceWithTaxSub(
              OBS=5 
              KEEP=make model invoice msrp msrpWithTax); 
      RUN;

   /* VOWELS (count) */
   DATA work.vowels(KEEP=product_group vowelCount);
     /* Using the products names from the first quarter 1991 sales 
        of children sports apparel, produce a data set with the 
        count of the number of vowels in the product name using 
        the getVowelCount FCMP function.                             */
     SET sashelp.orsales(KEEP=quarter product_line 
                              product_category product_group);
     WHERE quarter="1999Q1"        AND 
           product_line="Children" AND 
           product_category="Children Sports";
     vowelCount=getVowelCount(product_group);
     RUN;

   PROC PRINT DATA=work.vowels;
      RUN;

   /* VOWELS (statement) */
   DATA work.vowelsStmt(KEEP=product_group vowelCountStmt);
     /* Using the products names from the first quarter 1991 sales 
        of children sports apparel, produce a data set with 
        sentences noting the product name and the count of the 
        number of vowels in the product name using the 
        getVowelCountStmt FCMP function.                             */
     SET sashelp.orsales(KEEP=quarter product_line 
                              product_category product_group);
     WHERE quarter="1999Q1"        AND 
           product_line="Children" AND 
           product_category="Children Sports";
     vowelCountStmt=getVowelCountStmt(product_group);
     RUN;

   PROC PRINT DATA=work.vowelsStmt(OBS=5);
      RUN;

   /* List available functions/subroutines */
   PROC FCMP INLIB=billsLib.paper LISTFUNCS;
      RUN;

   /* Delete a function and a subroutine definition from the FCMP 
	  data set.                                                      */
   PROC FCMP INLIB=billsLib.paper OUTLIB=billsLib.paper.delete;
       DELETEFUNC priceWithTax_test1;
       DELETESUBR priceWithTax_test2;
       RUN;

   /* Temperature averages */
   DATA work.monthTemps;
      /* Create a data set of 4 times daily temperature reading over 
         a monthly period.                                           */
      INPUT AM6 PM12 PM6 AM12;
      DATALINES;
      28 32 39 23
      17 24 29 14
      18 25 30 15
      20 27 32 17
      14 16 18  7
      17 20 23 13
      18 21 24 14
      33 25 14 11
      27 31 38 22
      32 24 13 10
      25 33 45 22
      45 51 56 44
      38 42 43 36
      39 43 44 37
      40 46 48 41
      46 52 57 45
      28 32 40 21
      24 32 44 21
      29 33 41 22
      26 34 47 24
      23 34 41 29
      24 35 42 30
      23 28 32 34
      32 33 34 32
      34 37 40 39
      19 26 31 16
      18 27 36 33
       7 11 14  9
      15 17 19  8
       8 26 33 21
      ;;;
  
   RUN;

   PROC PRINT DATA=work.monthTemps(OBS=5); 
      RUN;

   DATA work.monthAvgs;
      /* Produce a data set that has the average temperatures at 4 
         times during the day over the course of a month.            */
      AM6Avg  = averageTempAtTime("work.monthTemps", 1);
      PM12Avg = averageTempAtTime("work.monthTemps", 2);
      PM6Avg  = averageTempAtTime("work.monthTemps", 3);
      AM12Avg = averageTempAtTime("work.monthTemps", 4);
   RUN;

   PROC PRINT DATA=work.monthAvgs; 
      RUN;

   /* Run macro */
   %MACRO jsonFileCreate();
      /* SAS Macro code called from the writeJsonFile FCMP function. 
         The macro will convert the quoted macro variables created
         by the Special FCMP function RUN_MACRO. The macro will
         create a JSON file be calling the JSON procedure.           */
      %LOCAL/READONLY inDsName   = %SYSFUNC(DEQUOTE(&dsName));
      %LOCAL/READONLY inAge      = %SYSFUNC(DEQUOTE(&age));
      %LOCAL/READONLY outputFile = "./age_&inAge..json";

	  /* Call JSON procedure to create JSON file.                    */
      PROC JSON OUT=&outputFile NOSASTAGS;
        EXPORT &inDsName (WHERE=(Age EQ &inAge));
        QUIT;

	  /* Error checking.                                             */
      %LET procRc = &syserr;
      %IF (&procRc EQ 0)
          %THEN %PUT Created file &outputFile;
      %MEND jsonFileCreate;

   %LET sortedAgeDataSetName=WORK.sortedAges;

   PROC SORT DATA=sashelp.class 
             OUT=&sortedAgeDataSetName;
	  /* For the BY statement to work, the data set must be sorted.  */
      BY Age;
      RUN;

   DATA _NULL_;
     /* For each unique age in the data set, produce a JSON file 
        that contains the information on the students that are of 
        the specified age. */
     SET &sortedAgeDataSetName;
     BY Age;
     IF (first.age)
       THEN DO;
	      /* Call the FCMP wrapper function that produces the JSON 
	         data file for the specified age value.                  */
          rc = writeJsonFile("&sortedAgeDataSetName", age);
		  /* Error checking */
          IF (rc NE 0)
            THEN ABORT;
          END;
     RUN;

   /* Recursion (FCMP function) */
   DATA work.facts;
      /* Call the FCMP function to compute the factorial.            */  
     DO number=5 TO 9;
        factorial=factorial(number);
        OUTPUT;
        END;
     RUN;

   PROC PRINT DATA=work.facts;
     RUN;

   /* Recursion (external C function) */
   /* Note that the CMPLIB= option needs both the location of the 
	 FCMP function definition and the PROTO interface definition.    */
   OPTION CMPLIB=(billsLib.paper billsLib.proto);

   DATA work.factsProto;
      /* Call the FCMP wrapper function that calls the external C 
         function to compute the factorial.                          */  
      DO number=5 TO 9;
         factorial=factorialProto(number);
         OUTPUT;
         END;
      RUN;

   PROC PRINT DATA=work.factsProto;
     RUN;

   /* Hash object */
   DATA work.possCars;
      /* Create a data set of the models of cars of interest.        */
      INPUT Model $40.;
      DATALINES;
         Element LX
         Malibu Maxx LS
         Rio Cinco
         Outback
         Rodeo S
         Matrix XR
         Xterra XE V6
         Murano SL
         Jetta GL
         Passat GLS 1.8T
         4Runner SR5 V6
         V40
         ;;;
      RUN;

   PROC PRINT DATA=work.possCars; 
      RUN;

   DATA work.carDetails;
      /* Using the WHERE statement IN operator, create a data set 
         with the information of the cars of interest.               */
      SET sashelp.cars;
      WHERE (STRIP(model) IN ("Element LX", "Malibu Maxx LS", 
                              "Rio Cinco", "Outback", "Rodeo S",
                              "Matrix XR", "Xterra XE V6",
                              "Murano SL", "Jetta GL", 
                              "Passat GLS 1.8T", 
                              "4Runner SR5 V6", "V40"));
   RUN;

   PROC PRINT DATA=work.carDetails(KEEP=make model type origin 
                                        drivetrain msrp invoice); 
      RUN;

   DATA work.carDetails;
      /* Using a FCMP function to replace the WHERE statement IN 
         operator, create a data set with the information of the 
         cars of interest.                                           */
      SET sashelp.cars;
      WHERE isPossibleCar(STRIP(model));
      RUN;

   PROC PRINT DATA=work.carDetails(KEEP=make model type origin 
                                        drivetrain msrp invoice); 
      RUN;
